@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- encoding: utf-8 -*-
 
-# Copyright (c) 2002-2020 "Neo4j,"
+# Copyright (c) "Neo4j"
 # Neo4j Sweden AB [http://neo4j.com]
 #
 # This file is part of Neo4j.
@@ -18,112 +18,119 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from collections import deque
+from enum import Enum
+from logging import getLogger
 from ssl import SSLSocket
-from time import perf_counter
-from neo4j.api import (
-    Version,
-    READ_ACCESS,
-    DEFAULT_DATABASE,
-)
-from neo4j.io._common import (
-    Inbox,
-    Outbox,
-    Response,
-    InitResponse,
-    CommitResponse,
-)
-from neo4j.meta import get_user_agent
-from neo4j.exceptions import (
-    AuthError,
-    ServiceUnavailable,
-    DatabaseUnavailable,
-    NotALeader,
-    ForbiddenOnReadOnlyDatabase,
-    SessionExpired,
-    ConfigurationError,
-)
+
 from neo4j._exceptions import (
-    BoltIncompleteCommitError,
+    BoltError,
     BoltProtocolError,
 )
-from neo4j.packstream import (
-    Unpacker,
-    Packer,
+from neo4j.api import (
+    READ_ACCESS,
+    Version,
+)
+from neo4j.exceptions import (
+    ConfigurationError,
+    DatabaseUnavailable,
+    DriverError,
+    ForbiddenOnReadOnlyDatabase,
+    Neo4jError,
+    NotALeader,
+    ServiceUnavailable,
 )
 from neo4j.io import (
     Bolt,
-    BoltPool,
+    check_supported_server_product,
+)
+from neo4j.io._common import (
+    CommitResponse,
+    InitResponse,
+    Response,
 )
-from neo4j.api import ServerInfo
-from neo4j.addressing import Address
 
-from logging import getLogger
+
 log = getLogger("neo4j")
 
 
+class ServerStates(Enum):
+    CONNECTED = "CONNECTED"
+    READY = "READY"
+    STREAMING = "STREAMING"
+    TX_READY_OR_TX_STREAMING = "TX_READY||TX_STREAMING"
+    FAILED = "FAILED"
+
+
+class ServerStateManager:
+    _STATE_TRANSITIONS = {
+        ServerStates.CONNECTED: {
+            "hello": ServerStates.READY,
+        },
+        ServerStates.READY: {
+            "run": ServerStates.STREAMING,
+            "begin": ServerStates.TX_READY_OR_TX_STREAMING,
+        },
+        ServerStates.STREAMING: {
+            "pull": ServerStates.READY,
+            "discard": ServerStates.READY,
+            "reset": ServerStates.READY,
+        },
+        ServerStates.TX_READY_OR_TX_STREAMING: {
+            "commit": ServerStates.READY,
+            "rollback": ServerStates.READY,
+            "reset": ServerStates.READY,
+        },
+        ServerStates.FAILED: {
+            "reset": ServerStates.READY,
+        }
+    }
+
+    def __init__(self, init_state, on_change=None):
+        self.state = init_state
+        self._on_change = on_change
+
+    def transition(self, message, metadata):
+        if metadata.get("has_more"):
+            return
+        state_before = self.state
+        self.state = self._STATE_TRANSITIONS\
+            .get(self.state, {})\
+            .get(message, self.state)
+        if state_before != self.state and callable(self._on_change):
+            self._on_change(state_before, self.state)
+
+
 class Bolt3(Bolt):
     """ Protocol handler for Bolt 3.
 
-    This is supported by Neo4j versions 3.5, 4.0, 4.1 and 4.2.
+    This is supported by Neo4j versions 3.5, 4.0, 4.1, 4.2, 4.3, and 4.4.
     """
 
     PROTOCOL_VERSION = Version(3, 0)
 
-    # The socket
-    in_use = False
+    supports_multiple_results = False
 
-    # The socket
-    _closed = False
+    supports_multiple_databases = False
 
-    # The socket
-    _defunct = False
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self._server_state_manager = ServerStateManager(
+            ServerStates.CONNECTED, on_change=self._on_server_state_change
+        )
 
-    #: The pool of which this connection is a member
-    pool = None
+    def _on_server_state_change(self, old_state, new_state):
+        log.debug("[#%04X]  State: %s > %s", self.local_port,
+                  old_state.name, new_state.name)
 
-    def __init__(self, unresolved_address, sock, max_connection_lifetime, *, auth=None, user_agent=None, routing_context=None):
-        self.unresolved_address = unresolved_address
-        self.socket = sock
-        self.server_info = ServerInfo(Address(sock.getpeername()), self.PROTOCOL_VERSION)
-        self.outbox = Outbox()
-        self.inbox = Inbox(self.socket, on_error=self._set_defunct)
-        self.packer = Packer(self.outbox)
-        self.unpacker = Unpacker(self.inbox)
-        self.responses = deque()
-        self._max_connection_lifetime = max_connection_lifetime
-        self._creation_timestamp = perf_counter()
-        self.supports_multiple_results = False
-        self.supports_multiple_databases = False
-        self._is_reset = True
-        self.routing_context = routing_context
-
-        # Determine the user agent
-        if user_agent:
-            self.user_agent = user_agent
-        else:
-            self.user_agent = get_user_agent()
-
-        # Determine auth details
-        if not auth:
-            self.auth_dict = {}
-        elif isinstance(auth, tuple) and 2 <= len(auth) <= 3:
-            from neo4j import Auth
-            self.auth_dict = vars(Auth("basic", *auth))
-        else:
-            try:
-                self.auth_dict = vars(auth)
-            except (KeyError, TypeError):
-                raise AuthError("Cannot determine auth details from %r" % auth)
-
-        # Check for missing password
-        try:
-            credentials = self.auth_dict["credentials"]
-        except KeyError:
-            pass
-        else:
-            if credentials is None:
-                raise AuthError("Password cannot be None")
+    @property
+    def is_reset(self):
+        # We can't be sure of the server's state if there are still pending
+        # responses. Unless the last message we sent was RESET. In that case
+        # the server state will always be READY when we're done.
+        if (self.responses and self.responses[-1]
+                and self.responses[-1].message == "reset"):
+            return True
+        return self._server_state_manager.state == ServerStates.READY
 
     @property
     def encrypted(self):
@@ -137,7 +144,7 @@ class Bolt3(Bolt):
     def local_port(self):
         try:
             return self.socket.getsockname()[1]
-        except IOError:
+        except OSError:
             return 0
 
     def get_base_headers(self):
@@ -153,13 +160,63 @@ class Bolt3(Bolt):
             logged_headers["credentials"] = "*******"
         log.debug("[#%04X]  C: HELLO %r", self.local_port, logged_headers)
         self._append(b"\x01", (headers,),
-                     response=InitResponse(self, on_success=self.server_info._update_metadata))
+                     response=InitResponse(self, "hello",
+                                           on_success=self.server_info.update))
         self.send_all()
         self.fetch_all()
+        check_supported_server_product(self.server_info.agent)
 
-    def run(self, query, parameters=None, mode=None, bookmarks=None, metadata=None, timeout=None, db=None, **handlers):
+    def route(self, database=None, imp_user=None, bookmarks=None):
+        if database is not None:
+            raise ConfigurationError(
+                "Database name parameter for selecting database is not "
+                "supported in Bolt Protocol {!r}. Database name {!r}. "
+                "Server Agent {!r}".format(
+                    self.PROTOCOL_VERSION, database, self.server_info.agent
+                )
+            )
+        if imp_user is not None:
+            raise ConfigurationError(
+                "Impersonation is not supported in Bolt Protocol {!r}. "
+                "Trying to impersonate {!r}.".format(
+                    self.PROTOCOL_VERSION, imp_user
+                )
+            )
+
+        metadata = {}
+        records = []
+
+        # Ignoring database and bookmarks because there is no multi-db support.
+        # The bookmarks are only relevant for making sure a previously created
+        # db exists before querying a routing table for it.
+        self.run(
+            "CALL dbms.cluster.routing.getRoutingTable($context)",  # This is an internal procedure call. Only available if the Neo4j 3.5 is setup with clustering.
+            {"context": self.routing_context},
+            mode="r",                                               # Bolt Protocol Version(3, 0) supports mode="r"
+            on_success=metadata.update
+        )
+        self.pull(on_success=metadata.update, on_records=records.extend)
+        self.send_all()
+        self.fetch_all()
+        routing_info = [dict(zip(metadata.get("fields", ()), values)) for values in records]
+        return routing_info
+
+    def run(self, query, parameters=None, mode=None, bookmarks=None,
+            metadata=None, timeout=None, db=None, imp_user=None, **handlers):
         if db is not None:
-            raise ConfigurationError("Database name parameter for selecting database is not supported in Bolt Protocol {!r}. Database name {!r}.".format(Bolt3.PROTOCOL_VERSION, db))
+            raise ConfigurationError(
+                "Database name parameter for selecting database is not "
+                "supported in Bolt Protocol {!r}. Database name {!r}.".format(
+                    self.PROTOCOL_VERSION, db
+                )
+            )
+        if imp_user is not None:
+            raise ConfigurationError(
+                "Impersonation is not supported in Bolt Protocol {!r}. "
+                "Trying to impersonate {!r}.".format(
+                    self.PROTOCOL_VERSION, imp_user
+                )
+            )
         if not parameters:
             parameters = {}
         extra = {}
@@ -183,36 +240,37 @@ class Bolt3(Bolt):
         fields = (query, parameters, extra)
         log.debug("[#%04X]  C: RUN %s", self.local_port, " ".join(map(repr, fields)))
         if query.upper() == u"COMMIT":
-            self._append(b"\x10", fields, CommitResponse(self, **handlers))
+            self._append(b"\x10", fields, CommitResponse(self, "run",
+                                                         **handlers))
         else:
-            self._append(b"\x10", fields, Response(self, **handlers))
-        self._is_reset = False
-
-    def run_get_routing_table(self, on_success, on_failure, database=DEFAULT_DATABASE):
-        if database != DEFAULT_DATABASE:
-            raise ConfigurationError("Database name parameter for selecting database is not supported in Bolt Protocol {!r}. Database name {!r}. Server Agent {!r}.".format(Bolt3.PROTOCOL_VERSION, database, self.server_info.agent))
-        self.run(
-            "CALL dbms.cluster.routing.getRoutingTable($context)",  # This is an internal procedure call. Only available if the Neo4j 3.5 is setup with clustering.
-            {"context": self.routing_context},
-            mode="r",                                               # Bolt Protocol Version(3, 0) supports mode="r"
-            on_success=on_success,
-            on_failure=on_failure,
-        )
+            self._append(b"\x10", fields, Response(self, "run", **handlers))
 
     def discard(self, n=-1, qid=-1, **handlers):
         # Just ignore n and qid, it is not supported in the Bolt 3 Protocol.
         log.debug("[#%04X]  C: DISCARD_ALL", self.local_port)
-        self._append(b"\x2F", (), Response(self, **handlers))
+        self._append(b"\x2F", (), Response(self, "discard", **handlers))
 
     def pull(self, n=-1, qid=-1, **handlers):
         # Just ignore n and qid, it is not supported in the Bolt 3 Protocol.
         log.debug("[#%04X]  C: PULL_ALL", self.local_port)
-        self._append(b"\x3F", (), Response(self, **handlers))
-        self._is_reset = False
+        self._append(b"\x3F", (), Response(self, "pull", **handlers))
 
-    def begin(self, mode=None, bookmarks=None, metadata=None, timeout=None, db=None, **handlers):
+    def begin(self, mode=None, bookmarks=None, metadata=None, timeout=None,
+              db=None, imp_user=None, **handlers):
         if db is not None:
-            raise ConfigurationError("Database name parameter for selecting database is not supported in Bolt Protocol {!r}. Database name {!r}.".format(Bolt3.PROTOCOL_VERSION, db))
+            raise ConfigurationError(
+                "Database name parameter for selecting database is not "
+                "supported in Bolt Protocol {!r}. Database name {!r}.".format(
+                    self.PROTOCOL_VERSION, db
+                )
+            )
+        if imp_user is not None:
+            raise ConfigurationError(
+                "Impersonation is not supported in Bolt Protocol {!r}. "
+                "Trying to impersonate {!r}.".format(
+                    self.PROTOCOL_VERSION, imp_user
+                )
+            )
         extra = {}
         if mode in (READ_ACCESS, "r"):
             extra["mode"] = "r"  # It will default to mode "w" if nothing is specified
@@ -232,28 +290,15 @@ class Bolt3(Bolt):
             except TypeError:
                 raise TypeError("Timeout must be specified as a number of seconds")
         log.debug("[#%04X]  C: BEGIN %r", self.local_port, extra)
-        self._append(b"\x11", (extra,), Response(self, **handlers))
-        self._is_reset = False
+        self._append(b"\x11", (extra,), Response(self, "begin", **handlers))
 
     def commit(self, **handlers):
         log.debug("[#%04X]  C: COMMIT", self.local_port)
-        self._append(b"\x12", (), CommitResponse(self, **handlers))
+        self._append(b"\x12", (), CommitResponse(self, "commit", **handlers))
 
     def rollback(self, **handlers):
         log.debug("[#%04X]  C: ROLLBACK", self.local_port)
-        self._append(b"\x13", (), Response(self, **handlers))
-
-    def _append(self, signature, fields=(), response=None):
-        """ Add a message to the outgoing queue.
-
-        :arg signature: the signature of the message
-        :arg fields: the fields of the message as a tuple
-        :arg response: a response object to handle callbacks
-        """
-        self.packer.pack_struct(signature, fields)
-        self.outbox.chunk()
-        self.outbox.chunk()
-        self.responses.append(response)
+        self._append(b"\x13", (), Response(self, "rollback", **handlers))
 
     def reset(self):
         """ Add a RESET message to the outgoing queue, send
@@ -264,42 +309,12 @@ class Bolt3(Bolt):
             raise BoltProtocolError("RESET failed %r" % metadata, address=self.unresolved_address)
 
         log.debug("[#%04X]  C: RESET", self.local_port)
-        self._append(b"\x0F", response=Response(self, on_failure=fail))
+        self._append(b"\x0F", response=Response(self, "reset", on_failure=fail))
         self.send_all()
         self.fetch_all()
-        self._is_reset = True
-
-    def _send_all(self):
-        data = self.outbox.view()
-        if data:
-            self.socket.sendall(data)
-            self.outbox.clear()
-
-    def send_all(self):
-        """ Send all queued messages to the server.
-        """
-        if self.closed():
-            raise ServiceUnavailable("Failed to write to closed connection {!r} ({!r})".format(
-                self.unresolved_address, self.server_info.address))
-
-        if self.defunct():
-            raise ServiceUnavailable("Failed to write to defunct connection {!r} ({!r})".format(
-                self.unresolved_address, self.server_info.address))
-
-        try:
-            self._send_all()
-        except (IOError, OSError) as error:
-            log.error("Failed to write data to connection "
-                      "{!r} ({!r}); ({!r})".
-                      format(self.unresolved_address,
-                             self.server_info.address,
-                             "; ".join(map(repr, error.args))))
-            if self.pool:
-                self.pool.deactivate(address=self.unresolved_address)
-            raise
 
     def fetch_message(self):
-        """ Receive at least one message from the server, if available.
+        """ Receive at most one message from the server, if available.
 
         :return: 2-tuple of number of detail messages and number of summary
                  messages fetched
@@ -316,17 +331,7 @@ class Bolt3(Bolt):
             return 0, 0
 
         # Receive exactly one message
-        try:
-            details, summary_signature, summary_metadata = next(self.inbox)
-        except (IOError, OSError) as error:
-            log.error("Failed to read data from connection "
-                      "{!r} ({!r}); ({!r})".
-                      format(self.unresolved_address,
-                             self.server_info.address,
-                             "; ".join(map(repr, error.args))))
-            if self.pool:
-                self.pool.deactivate(address=self.unresolved_address)
-            raise
+        details, summary_signature, summary_metadata = next(self.inbox)
 
         if details:
             log.debug("[#%04X]  S: RECORD * %d", self.local_port, len(details))  # Do not log any data
@@ -339,12 +344,15 @@ class Bolt3(Bolt):
         response.complete = True
         if summary_signature == b"\x70":
             log.debug("[#%04X]  S: SUCCESS %r", self.local_port, summary_metadata)
+            self._server_state_manager.transition(response.message,
+                                                  summary_metadata)
             response.on_success(summary_metadata or {})
         elif summary_signature == b"\x7E":
             log.debug("[#%04X]  S: IGNORED", self.local_port)
             response.on_ignored(summary_metadata or {})
         elif summary_signature == b"\x7F":
             log.debug("[#%04X]  S: FAILURE %r", self.local_port, summary_metadata)
+            self._server_state_manager.state = ServerStates.FAILED
             try:
                 response.on_failure(summary_metadata or {})
             except (ServiceUnavailable, DatabaseUnavailable):
@@ -355,67 +363,15 @@ class Bolt3(Bolt):
                 if self.pool:
                     self.pool.on_write_failure(address=self.unresolved_address),
                 raise
+            except Neo4jError as e:
+                if self.pool and e.invalidates_all_connections():
+                    self.pool.mark_all_stale()
+                raise
         else:
             raise BoltProtocolError("Unexpected response message with signature %02X" % summary_signature, address=self.unresolved_address)
 
         return len(details), 1
 
-    def _set_defunct(self, error=None):
-        direct_driver = isinstance(self.pool, BoltPool)
-
-        message = ("Failed to read from defunct connection {!r} ({!r})".format(
-            self.unresolved_address, self.server_info.address))
-
-        if error:
-            log.error(str(error))
-        log.error(message)
-        # We were attempting to receive data but the connection
-        # has unexpectedly terminated. So, we need to close the
-        # connection from the client side, and remove the address
-        # from the connection pool.
-        self._defunct = True
-        self.close()
-        if self.pool:
-            self.pool.deactivate(address=self.unresolved_address)
-        # Iterate through the outstanding responses, and if any correspond
-        # to COMMIT requests then raise an error to signal that we are
-        # unable to confirm that the COMMIT completed successfully.
-        for response in self.responses:
-            if isinstance(response, CommitResponse):
-                if error:
-                    raise BoltIncompleteCommitError(message, address=None) from error
-                else:
-                    raise BoltIncompleteCommitError(message, address=None)
-
-        if direct_driver:
-            if error:
-                raise ServiceUnavailable(message) from error
-            else:
-                raise ServiceUnavailable(message)
-        else:
-            if error:
-                raise SessionExpired(message) from error
-            else:
-                raise SessionExpired(message)
-
-    def timedout(self):
-        return 0 <= self._max_connection_lifetime <= perf_counter() - self._creation_timestamp
-
-    def fetch_all(self):
-        """ Fetch all outstanding messages.
-
-        :return: 2-tuple of number of detail messages and number of summary
-                 messages fetched
-        """
-        detail_count = summary_count = 0
-        while self.responses:
-            response = self.responses[0]
-            while not response.complete:
-                detail_delta, summary_delta = self.fetch_message()
-                detail_count += detail_delta
-                summary_count += summary_delta
-        return detail_count, summary_count
-
     def close(self):
         """ Close the connection.
         """
@@ -425,12 +381,12 @@ class Bolt3(Bolt):
                 self._append(b"\x02", ())
                 try:
                     self._send_all()
-                except:
+                except (OSError, BoltError, DriverError):
                     pass
             log.debug("[#%04X]  C: <CLOSE>", self.local_port)
             try:
                 self.socket.close()
-            except IOError:
+            except OSError:
                 pass
             finally:
                 self._closed = True
